searchNodes=[{"ref":"Lab42.SimpleStateMachine.html","title":"Lab42.SimpleStateMachine","type":"module","doc":"A simple state machine. SimpleStateMachine is a minimalistic approach to write State Machines which operate on a list of inputs. The machine is defined by a map, mapping, transitions to each state, called transition_map. %{ start: [ transition, ... ], some_state: [transition, ...] } Format of a Transition {trigger, transition_fn, new_state} or, if the state is not supposed to change, {trigger, transition_fn} For each input in the list of inputs the machine will try to match the input with the trigger from the transition. If such a match is found the transition_fn is called with a %SimpleStateMachine.Match struct which will give access to the following fields: %SimpleStateMachine.Match{ input: &quot;The element from the input list&quot;, data: &quot;Injected value when the state machine is run, like an accumulator in Enum.reduce&quot;, matched: &quot;A value depending on what kind of trigger was used&quot;} The return value of the transition_fn will be injected into the Match struct&#39;s data: field for the next loop. Types of triggers Function triggers ...are the most versatile triggers, when a function trigger triggers on an input it returns an unfalsy value that is passed into the Match struct&#39;s matched: field. Regex triggers ...can, obviously, only be used with String inputs. The result of Regex.run(trigger, input) is passed into the Match struct&#39;s matched: field. true trigger ... matches always, true is passed into the matched: field. :end trigger ... does never match, however its associated transaction_fn is called, and its result will bet the result of the machine&#39;s run. See also the end: state below. Special States Two states are special in the sense that their names are fixed. :start state Is defined like any other state but is the machine&#39;s initial current_state. It is obviously necessarily present in the transition_map. :end state Can only have one transition_fn which is invoked at the end of the input list. Reserved States :halt state No transition definitions for this state will ever be read. If the current_state of the machine becomes the :halt state, it stops and returns the Match struct&#39;s data: field. No :end state or trigger treatment is performed. :error state and states starting with :_ Reserved for future use. Some Detailed Examples Let&#39;s start with a single state machine. iex(0)&gt; parse_and_add = fn(string, data) -&gt; ...(0)&gt; {n, _} = Integer.parse(string) ...(0)&gt; %{data|sum: data.sum + n} end ...(0)&gt; add_error = fn(%{input: input, data: data}) -&gt; ...(0)&gt; %{data|errors: [input|data.errors]} end ...(0)&gt; states = %{ ...(0)&gt; start: [ ...(0)&gt; {~r(\\d+), fn %{matched: [d], data: data} -&gt; parse_and_add.(d, data) end}, ...(0)&gt; {true, add_error}, ...(0)&gt; ], ...(0)&gt; end: fn %{data: %{errors: errors, sum: sum}} -&gt; {sum, Enum.reverse(errors)} end } ...(0)&gt; run(~w{12 error 30 incorrect}, %{sum: 0, errors: []}, states) {42, ~w(error incorrect)} If the data is initially nil it needs not be passed into run and if the transaction_fn is a nop, it can be designated by nil. iex(1)&gt; states = %{ ...(1)&gt; start: [ ...(1)&gt; { ~r{(\\d+)}, fn %{matched: [_, d]} -&gt; d end, :halt }, ...(1)&gt; { true, nil } ]} ...(1)&gt; run(~w{ hello 42 84 }, states) &quot;42&quot; The difference between :halt and :end can be demonstrated with these slighly modified machines iex(2)&gt; sm1 = %{ ...(2)&gt; start: [ ...(2)&gt; { ~r{(\\d+)}, fn %{matched: [_, d]} -&gt; d end, :halt }, ...(2)&gt; { true, nil } ], ...(2)&gt; end: fn %{data: x} -&gt; {n, _} = Integer.parse(x); n end } ...(2)&gt; sm2 = %{ ...(2)&gt; start: [ ...(2)&gt; { ~r{(\\d+)}, fn %{matched: [_, d]} -&gt; d end, :end }, ...(2)&gt; { true, nil } ], ...(2)&gt; end: fn %{data: x} -&gt; {n, _} = Integer.parse(x); n end } ...(2)&gt; { run(~w{ hello 42 84 }, sm1), run(~w{ hello 42 84 }, sm2) } {&quot;42&quot;, 42} So far we have only seen Regex and true triggers, the next example uses function triggers iex(3)&gt; odd? = &amp;(rem(&amp;1, 2) == 1) ...(3)&gt; states = %{ ...(3)&gt; start: [ ...(3)&gt; {odd?, fn %{input: n, data: sum} -&gt; sum + n end}, ...(3)&gt; {true} ] } ...(3)&gt; run(1..6|&gt;Enum.into([]), 0, states) 9 Some might suggest that the {true} transition should be a default, but we prefer to raise an error if no transition matches iex(4)&gt; odd? = &amp;(rem(&amp;1, 2) == 1) ...(4)&gt; states = %{ ...(4)&gt; start: [ ...(4)&gt; {odd?, fn %{input: n, data: sum} -&gt; sum + n end} ]} ...(4)&gt; run(1..6|&gt;Enum.into([]), 0, states) ** (RuntimeError) No transition found in state :start, on input 2 An even more obvious exception is raised if a state has no transitions defined, that holds for the predefined :start state as for any other state. iex(5)&gt; states=%{} ...(5)&gt; run(~w[alpha beta], states) ** (RuntimeError) No transitions defined for state :start iex(6)&gt; states=%{ ...(6)&gt; start: [ ...(6)&gt; {true, nil, :second} ]} ...(6)&gt; run(~w[alpha beta], states) ** (RuntimeError) No transitions defined for state :second"},{"ref":"Lab42.SimpleStateMachine.html#run/3","title":"Lab42.SimpleStateMachine.run/3","type":"function","doc":""},{"ref":"Lab42.SimpleStateMachine.html#t:complete_transition_t/0","title":"Lab42.SimpleStateMachine.complete_transition_t/0","type":"type","doc":""},{"ref":"Lab42.SimpleStateMachine.html#t:function_t/0","title":"Lab42.SimpleStateMachine.function_t/0","type":"type","doc":""},{"ref":"Lab42.SimpleStateMachine.html#t:maybe/1","title":"Lab42.SimpleStateMachine.maybe/1","type":"type","doc":""},{"ref":"Lab42.SimpleStateMachine.html#t:state_t/0","title":"Lab42.SimpleStateMachine.state_t/0","type":"type","doc":""},{"ref":"Lab42.SimpleStateMachine.html#t:transition_fn_t/0","title":"Lab42.SimpleStateMachine.transition_fn_t/0","type":"type","doc":""},{"ref":"Lab42.SimpleStateMachine.html#t:transition_t/0","title":"Lab42.SimpleStateMachine.transition_t/0","type":"type","doc":""},{"ref":"Lab42.SimpleStateMachine.html#t:trigger_t/0","title":"Lab42.SimpleStateMachine.trigger_t/0","type":"type","doc":""},{"ref":"Lab42.SimpleStateMachine.Data.html","title":"Lab42.SimpleStateMachine.Data","type":"module","doc":"Represents the matched input and all data to be passed between state machine loops"},{"ref":"Lab42.SimpleStateMachine.Data.html#from_data/1","title":"Lab42.SimpleStateMachine.Data.from_data/1","type":"function","doc":""},{"ref":"Lab42.SimpleStateMachine.Data.html#t:t/0","title":"Lab42.SimpleStateMachine.Data.t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html","title":"Lab42.StateMachine","type":"module","doc":"Lab42.StateMachine Synopsis A simple State Machine operating on a list of input values, a map of transition definitions and an accumulator. What is the Transition Definitions Map? It maps each state to a list of transitions. Each transition is of the following format, as defined by type transition_t {trigger, transformer, updater, new_state} State, did you say state? Yes I did, and what I mean by it is that the State Machine keeps track of its state (very surprising) and some data, passed into by the user. How do these Transitions transform the input? The State Machine processes its input in a loop in which, depending on the current input element, designated as input and the current state, designated as current_state a transition is triggered. Data is passed in like an accumulator in Enum.reduce it is designated as data. The triggered transition will define the current_state of the next loop, if any, and perform actions defining what goes to the output and how data changes. As one might guess these actions are performed by the transformer and the updater of the triggered transition. So what is the contract? Starting the whole thing. Create the Transition Defintions Map, knowing that the State Machine will start with current_state equal to :start, yes I know, naming is difficult ;). Then run the whole thing like that: Lab42.StateMachine.run(input, my_data, %{}) The empty map passed in will cause havoc though iex(0)&gt; input = ~w(alpha) ...(0)&gt; my_data = %{n: 42} ...(0)&gt; Lab42.StateMachine.run(input, my_data, %{}) {:error, &quot;No transitions found for current state&quot;, :start} Specifying Transitions A minimal example might be a line counter iex(1)&gt; input = ~w{alpha beta} ...(1)&gt; count = 0 ...(1)&gt; states = %{ ...(1)&gt; start: [ {true, fn {_, line} -&gt; line end, fn count, _ -&gt; count + 1 end, :start} ] ...(1)&gt; } ...(1)&gt; run(input, count, states) {:start, ~w{alpha beta}, 2} N.B. That the true trigger always matches, other triggers are either regular expressions, which are used as follows: Regex.match(trigger, input), or functions, which are used as one might easily guess, thusly: trigger.(input) One could argue that a default behavior of copying the current input to the output might be convenient, but that might lead to difficulties in debugging state machines. (Maybe in later versions with an option for run?) On the same token some StateMachines, like the counter above collect the output without needing it, although we will learn below how to avoid this, a global option to not collect will make the Transitions Map more concise. Therefore the following will happen iex(2)&gt; input = ~w{alpha beta} ...(2)&gt; states = %{ ...(2)&gt; start: [] ...(2)&gt; } ...(2)&gt; run(input, nil, states) {:error, &quot;No trigger matched the current input \\&quot;alpha\\&quot;&quot;, :start} Let us return to the correctly working example, let us simplify that rather expressive transition iex(3)&gt; input = ~w{alpha beta} ...(3)&gt; states = %{ ...(3)&gt; start: [ {true, :id, fn count, _ -&gt; count + 1 end, :start} ] ...(3)&gt; } ...(3)&gt; run(input, 0, states) {:start, ~w(alpha beta), 2} So we can use a shortcut for copying the input to the output, that is better already, but still, why create the output that is not needed, let us use the atom form of the transformer function iex(4)&gt; input = ~w{alpha beta} ...(4)&gt; states = %{ ...(4)&gt; start: [ {true, fn _ -&gt; :ignore end, fn count, _ -&gt; count + 1 end, :start} ] ...(4)&gt; } ...(4)&gt; run(input, 0, states) {:start, ~w(), 2} As there is a shortcut for :id so there is one for :ignore iex(5)&gt; input = ~w{alpha beta} ...(5)&gt; states = %{ ...(5)&gt; start: [ {true, :ignore, fn count, _ -&gt; count + 1 end, :start} ] ...(5)&gt; } ...(5)&gt; run(input, 0, states) {:start, ~w(), 2} But what if we want to have :ignore in the output? Let us assume that we want to replace all &quot;alphas&quot; with :ignore. We can use the tuple form of the transformer in this case. This example also demonstrates that we do not need to specify the updater and the new state if these do not change and that even pushing the input to the output can be omitted, therefore the transitions {true, :id} and {true} have the same semantics. And a third simplifaction is that we can omit to pass nil as data, but note that it will present in the result. iex(6)&gt; input = ~w{alpha beta alpha} ...(6)&gt; states = %{ ...(6)&gt; start: [ ...(6)&gt; {~r{alpha}, fn _ -&gt; {:push, :ignore} end}, ...(6)&gt; {true}] } ...(6)&gt; run(input, states) {:start, [:ignore, &quot;beta&quot;, :ignore], nil} Now let us look at a complex example, that will use the following features, not yet explored: Stopping with the :halt atom form of the transformer, by its shortcut Stopping with the {:halt, value} tuple form of the transformer Creating a function to rerun with the same states Using constant functions Using function triggers iex(7)&gt; summer = fn data, {_, input} -&gt; %{data|sum: data.sum + input} end ...(7)&gt; states = %{ ...(7)&gt; start: [ ...(7)&gt; {&amp;(&amp;1&gt;8), :halt}, ...(7)&gt; {&amp;(&amp;1&lt;1), fn {_, input} -&gt; {:halt, input} end, constant(:negative)}, ...(7)&gt; {&amp;(rem(&amp;1,2)==0), push_constant(:even), summer, :even}, ...(7)&gt; {true} ], ...(7)&gt; even: [ ...(7)&gt; {&amp;(rem(&amp;1,2)==0), :halt, constant(:more_evens)}, ...(7)&gt; {true, :id, summer } ...(7)&gt; ] } ...(7)&gt; state_machine = make_machine(%{sum: 0}, states) ...(7)&gt; [ ...(7)&gt; state_machine.(1..10|&gt;Enum.into([])), ...(7)&gt; state_machine.([1, 2, 3]), ...(7)&gt; state_machine.([1, 9]), ...(7)&gt; state_machine.([1, -1]), ...(7)&gt; state_machine.([1, 3, 5]) ] [ {:even, [1, :even, 3], :more_evens}, {:even, [1, :even, 3], %{sum: 5}}, {:start, [1], %{sum: 0}}, {:start, [1, -1], :negative}, {:start, [1, 3, 5], %{sum: 0}} ] Summary of Shortcuts and Transformer Semantics As we have seen in the examples above, the StateMachine is driven by the Transitions Map and the result of the transformer function. The returned data value is influenced by the result of the updater function. In order to write shorter Transition Maps transitions can be shortened and symbolic function names can be used. Let us see how transitions are normalized by the StateMachine before they are executed. SpecifiedNormalizeRemarks {trigger, function, function, state}same valuealready normalized {trigger, function, function}{trigger, function, function, current_state}normalization is done during tuntime, thusly current_state is known {trigger, function }{trigger, function, identity2, current_state}given the interface of updater identity2 is defined as fn d, _ -&gt; d end |{trigger} | {trigger, identity1, identity2, current_state} | identity1 is for the transformer and thusly fn {_, v} -&gt; v end As can be seen sometimes the defaults, are not available, because we need to provide the new_state in the transition. In that case the following symbols can be used for the transformer Symbolic TransformerExpanded to :idfn {_, v} -&gt; v end :ignorefn _ -&gt; :ignore end :haltfn _ -&gt; :halt end For the updater there is only :id which expands to fn d, _ -&gt; d end as mentioned above. In action that would give iex(8)&gt; states = %{ ...(8)&gt; start: [{true, :id, :id, :stop}], ...(8)&gt; stop: [{true, fn _ -&gt; {:halt, &quot;...&quot;} end, :id}] ...(8)&gt; } ...(8)&gt; run(~w[Headline Rest MoreRest], states) {:stop, ~w[Headline ...], nil}"},{"ref":"Lab42.StateMachine.html#constant/1","title":"Lab42.StateMachine.constant/1","type":"function","doc":"A helper creating an updater function that sets data to a constant value"},{"ref":"Lab42.StateMachine.html#make_machine/2","title":"Lab42.StateMachine.make_machine/2","type":"function","doc":"A convenience helper to run the same state machine on different inputs"},{"ref":"Lab42.StateMachine.html#push_constant/1","title":"Lab42.StateMachine.push_constant/1","type":"function","doc":"A helper creating a transformer function that pushes a constant value to the output"},{"ref":"Lab42.StateMachine.html#run/3","title":"Lab42.StateMachine.run/3","type":"function","doc":""},{"ref":"Lab42.StateMachine.html#t:empty_t/0","title":"Lab42.StateMachine.empty_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:error_t/0","title":"Lab42.StateMachine.error_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:incomplete_transition_t/0","title":"Lab42.StateMachine.incomplete_transition_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:match_t/0","title":"Lab42.StateMachine.match_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:maybe/1","title":"Lab42.StateMachine.maybe/1","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:normalized_transitions_t/0","title":"Lab42.StateMachine.normalized_transitions_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:ok_t/0","title":"Lab42.StateMachine.ok_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:result_t/0","title":"Lab42.StateMachine.result_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:state_t/0","title":"Lab42.StateMachine.state_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:symbolic_transformer_t/0","title":"Lab42.StateMachine.symbolic_transformer_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:symbolic_updater_t/0","title":"Lab42.StateMachine.symbolic_updater_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:transformer_t/0","title":"Lab42.StateMachine.transformer_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:transition_map_t/0","title":"Lab42.StateMachine.transition_map_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:transition_t/0","title":"Lab42.StateMachine.transition_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:trigger_t/0","title":"Lab42.StateMachine.trigger_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:updater_t/0","title":"Lab42.StateMachine.updater_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html","title":"Lab42.StateMachine.Runner","type":"module","doc":"Runs the state machine by finding and executing transitions"},{"ref":"Lab42.StateMachine.Runner.html#halt_transfo/1","title":"Lab42.StateMachine.Runner.halt_transfo/1","type":"function","doc":"Convenience transformer function to stop the state machine, can be used with the atom :halt"},{"ref":"Lab42.StateMachine.Runner.html#ident_transfo/1","title":"Lab42.StateMachine.Runner.ident_transfo/1","type":"function","doc":"Convenience function to push an input to the output without changing it, can be used with the atom :id in the transformer position of a transition, e.g. {~r{alpha}, :id, fn count, _ -&gt; count + 1 end}"},{"ref":"Lab42.StateMachine.Runner.html#ident_updater/2","title":"Lab42.StateMachine.Runner.ident_updater/2","type":"function","doc":"Convenience function to not change the data. It can be used with the atom :id in the updater position of a transition, e.g. {~r{alpha}, fn {_, line} -&gt; String.reverse(line), :id}"},{"ref":"Lab42.StateMachine.Runner.html#ignore_input/1","title":"Lab42.StateMachine.Runner.ignore_input/1","type":"function","doc":"Convenience function to ignore an input, it can be used with the atom :ignore in the transformer position of a transition, e.g. {~r{alpha}, :ignore, fn count, _ -&gt; count + 1 end}"},{"ref":"Lab42.StateMachine.Runner.html#t:empty_t/0","title":"Lab42.StateMachine.Runner.empty_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:error_t/0","title":"Lab42.StateMachine.Runner.error_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:incomplete_transition_t/0","title":"Lab42.StateMachine.Runner.incomplete_transition_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:match_t/0","title":"Lab42.StateMachine.Runner.match_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:maybe/1","title":"Lab42.StateMachine.Runner.maybe/1","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:normalized_transitions_t/0","title":"Lab42.StateMachine.Runner.normalized_transitions_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:ok_t/0","title":"Lab42.StateMachine.Runner.ok_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:result_t/0","title":"Lab42.StateMachine.Runner.result_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:state_t/0","title":"Lab42.StateMachine.Runner.state_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:symbolic_transformer_t/0","title":"Lab42.StateMachine.Runner.symbolic_transformer_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:symbolic_updater_t/0","title":"Lab42.StateMachine.Runner.symbolic_updater_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:transformer_t/0","title":"Lab42.StateMachine.Runner.transformer_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:transition_map_t/0","title":"Lab42.StateMachine.Runner.transition_map_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:transition_t/0","title":"Lab42.StateMachine.Runner.transition_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:trigger_t/0","title":"Lab42.StateMachine.Runner.trigger_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:updater_t/0","title":"Lab42.StateMachine.Runner.updater_t/0","type":"type","doc":""}]